package org.jvnet.hypercompass.runtime.compass.accessor;

import java.beans.Introspector;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.compass.core.CompassException;
import org.compass.core.accessor.Getter;
import org.compass.core.accessor.PropertyAccessor;
import org.compass.core.accessor.PropertyNotFoundException;
import org.compass.core.accessor.Setter;

public abstract class CheckingAccessor implements PropertyAccessor {

  private final PropertyAccessor corePropertyAccessor;

  public CheckingAccessor(PropertyAccessor corePropertyAccessor) {
    this.corePropertyAccessor = corePropertyAccessor;
  }

  public Getter getGetter(Class theClass, String name) throws PropertyNotFoundException {

    final Getter getter = corePropertyAccessor.getGetter(theClass, name);

    final Method isSetter = getIsSetter(theClass, name);

    if (isSetter == null) {
      return getter;
    }
    else {
      return new CheckingGetter(name, getter, isSetter);
    }
  }

  private Method getIsSetter(Class theClass, String name) {
    final Method[] methods = theClass.getMethods(); // TSI
    for (int i = 0; i < methods.length; i++) {
      // only carry on if the method has no parameters
      if (methods[i].getParameterTypes().length == 0) {
        String methodName = methods[i].getName();

        if (methodName.startsWith("isSet")) {
          String testStdMethod = Introspector.decapitalize(methodName.substring(5));
          String testOldMethod = methodName.substring(5);
          if ((testStdMethod.equals(name) || testOldMethod.equals(name))
              && (methods[i].getReturnType().equals(Boolean.class) || methods[i]
                  .getReturnType()
                  .equals(boolean.class))) {
            return methods[i];
          }
        }
      }
    }
    return null;
  }

  public static class CheckingGetter implements Getter {
    private final String name;
    private final Getter getter;
    private final Method isSetter;

    public boolean isSet(Object object) throws CompassException {
      try {
        return ((Boolean) isSetter.invoke(object, null)).booleanValue();
      }
      catch (IllegalArgumentException e) {
        throw new CompassException("Error checking field.", e);
      }
      catch (IllegalAccessException e) {
        throw new CompassException("Error checking field.", e);
      }
      catch (InvocationTargetException e) {
        throw new CompassException("Error checking field.", e);
      }
    }

    public CheckingGetter(String name, Getter getter, Method isSetter) {
      this.name = name;
      this.getter = getter;
      this.isSetter = isSetter;
    }
    
    public String getName() {
      return name;
    }

    public Object get(Object object) throws CompassException {
      if (isSet(object)) {
        return getter.get(object);
      }
      else {
        return null;
      }
    }

    public Method getMethod() {
      return null;
    }

    public String getMethodName() {
      return null;
    }

    public Class getReturnType() {
      return getter.getReturnType();
    }
  }

  public Setter getSetter(Class theClass, String name) throws PropertyNotFoundException {
    final Setter setter = corePropertyAccessor.getSetter(theClass, name);

    final Method unsetter = getUnsetter(theClass, name);

    if (unsetter == null) {
      return setter;
    }
    else {
      return new CheckingSetter(name, setter, unsetter);
    }
  }

  /**
   * @param theClass
   * @param name
   * @return
   */
  private Method getUnsetter(Class theClass, String name) {
    final Method[] methods = theClass.getMethods(); // TSI
    for (int i = 0; i < methods.length; i++) {
      if (methods[i].getParameterTypes().length == 0) {
        String methodName = methods[i].getName();

        if (methodName.startsWith("unset")) {
          String testStdMethod = Introspector.decapitalize(methodName.substring(5));
          String testOldMethod = methodName.substring(5);
          if ((testStdMethod.equals(name) || testOldMethod.equals(name))
              && (methods[i].getReturnType().equals(Void.class) || methods[i]
                  .getReturnType()
                  .equals(void.class))) {
            return methods[i];
          }
        }
      }
    }
    return null;
  }

  public static class CheckingSetter implements Setter {
    private final String name;
    private final Setter setter;
    private final Method unsetter;

    public void unset(Object object) throws CompassException {
      try {
        unsetter.invoke(object, null);
      }
      catch (IllegalArgumentException e) {
        throw new CompassException("Error unsetting field.", e);
      }
      catch (IllegalAccessException e) {
        throw new CompassException("Error unsetting field.", e);
      }
      catch (InvocationTargetException e) {
        throw new CompassException("Error unsetting field.", e);
      }
    }

    public CheckingSetter(final String name, Setter setter, Method unsetter) {
      this.name = name;
      this.setter = setter;
      this.unsetter = unsetter;
    }

    public void set(Object object, Object value)
        throws CompassException {

      if (value != null) {
        setter.set(object, value);
      }
      else {
        unset(object);
      }
    }
    
    public String getName() {
      return name;
    }

    public Method getMethod() {
      return null;
    }

    public String getMethodName() {
      return null;
    }
  }
}
